param(
  [switch] $SkipParamInput = $false,
  [switch] $DownloadOnly = $false
)

if([System.IntPtr]::Size -eq 4)
{
  $message = "The 64-bit version of PowerShell is required to run this script."
  $message | Out-File -Append log.txt
  Write-Host $message
  exit 1
}

if (!([Security.Principal.WindowsPrincipal][Security.Principal.WindowsIdentity]::GetCurrent()).IsInRole([Security.Principal.WindowsBuiltInRole] "Administrator")) {
  $extendRunCommand = ""
  if ($SkipParamInput){
    $extendRunCommand += " -SkipParamInput"
  }
  if ($DownloadOnly){
    $extendRunCommand += " -DownloadOnly"
  }
  $cmd = $PSCommandPath + $extendRunCommand

  Write-Host "The script is NOT running as Administrator, restarting PowerShell as Administrator..."
  $cmd = $MyInvocation.MyCommand.Path + $extendRunCommand
  $arguments = "-NoProfile -NoExit -ExecutionPolicy Bypass -Command ""& {$cmd} """
  Start-Process "$psHome\powershell.exe" -Verb Runas -ArgumentList $arguments -ErrorAction 'stop'
  exit 1
}
cd $PSScriptRoot
clear;

if (Test-Path ".\config.json"){
  $settings = (Get-Content ".\config.json" -Raw) | ConvertFrom-Json
  $constants = (Get-Content ".\scripts\settings.json" -Raw) | ConvertFrom-Json
  $settings.IISAuthentication = ($settings.IISAuthentication).ToLower()
}
else{
  Write-Host "The configuration file config.json was not found in the present working directory."
}

if (Test-Path ".\tempconfig.json"){
  Remove-Item ".\tempconfig.json"
}

#Get present working directory
$pwd = (Get-Item -Path ".\" -Verbose).FullName
$package = $pwd + "\package"
$wc = New-Object Net.WebClient
$app = (New-Object -com shell.application)

$date = (Get-Date).ToUniversalTime()
$message = "Installation start: $date --- v" + $constants.PSInstallVersion | Out-File -Append log.txt
#Checking and creating directory paths
if (!(Test-Path $package)){
  $message = "0. Creating package folder" | Out-File -Append log.txt
  New-Item $package -type directory | Out-File -Append log.txt
}

function CheckDbSchema {
  try{
    $sqlCheckSchema = "SELECT COUNT(*) AS tables_found_count FROM INFORMATION_SCHEMA.TABLES WHERE TABLE_NAME IN ('DIRECTORY', 'FAVORITE', 'FAVORITE_TYPE', 'GROUPS', 'PROFILE', 'USERS', 'USERS_GROUPS', 'USERS_PREFS', 'USERS_PROFILE', 'USERS_VIEWS', 'USERS_VIEWS_MAPTO', 'WFACTIVITY', 'WFACTIVITY_ASSIGN_TYPE', 'WFACTIVITY_INST', 'WFACTIVITY_INST_MAPTO', 'WFACTIVITY_INST_NOTIFY_OVERDUE', 'WFACTIVITY_INST_RELDATA', 'WFACTIVITY_NOTIFY', 'WFACTIVITY_PARAM', 'WFACTIVITY_TYPE', 'WFAPPLICATION', 'WFAPPLICATION_PARAM', 'WFAPPLICATION_TYPE', 'WFCATEGORY', 'WFCOUNTER', 'WFDATASET', 'WFDATASET_VALUE', 'WFDATATYPE', 'WFDELEGATE', 'WFDURATION_UNIT', 'WFFOLDER', 'WFFOLDER_PERM', 'WFLIST', 'WFLIST_ITEMS', 'WFNOTIFY_TYPE', 'WFPARTICIPANT', 'WFPARTICIPANT_MAPTO', 'WFPARTICIPANT_ROLE', 'WFPARTICIPANT_TYPE', 'WFPERM_TYPE', 'WFPROCESS', 'WFPROCESS_CATEGORY', 'WFPROCESS_INST', 'WFPROCESS_INST_COMMENT', 'WFPROCESS_INST_RELDATA', 'WFPROCESS_PARTICIPANT', 'WFPROCESS_RELDATA', 'WFPROCESS_RELDATA_RESTRICT', 'WFPROCESS_STATUS', 'WFPROCESS_TRANSITION', 'WFRELDATA', 'WFSTATE', 'WFSUBSTATE')"
    $cmd = new-object system.data.sqlclient.sqlcommand($sqlCheckSchema, $sqlConnection)
    $tableCountCheck = ($constants.WFGDBTableCount -eq [convert]::ToInt32($cmd.ExecuteScalar(), 10))
    $sqlCheckConsts = "SELECT count(*) from FAVORITE_TYPE"
    $cmd = new-object system.data.sqlclient.sqlcommand($sqlCheckConsts, $sqlConnection)
    $constCountCheck = ($constants.WFGDBConstCount -eq [convert]::ToInt32($cmd.ExecuteScalar(), 10))
    if ($constCountCheck -And $tableCountCheck){
      $true
    }
    else{
      $false
    }
  }
  catch{
    $false
  }
}
#WELCOME MESSAGE***********************************************************************
Write-Host "`nWorkflowGen"$constants.WFGVersion "installation script"
Write-Host "`n`n`n`n"

if(!$DownloadOnly){
  if(!$SkipParamInput){
    Write-Host "`nFor each configuration parameter, press 'ENTER' to accept the current value or any other keys to change."
    Write-Host "Note: All parameters are required unless otherwise specified."
    Write-Host "`n"
    Write-Host -NoNewLine "License type:" $settings.LicenseType
    $key = $host.UI.RawUI.ReadKey("NoEcho,IncludeKeyDown")
    if ($key.VirtualKeyCode -ne 13){
      $settings.LicenseType = ""
      while (($settings.LicenseType).ToLower() -ne "trial" -And ($settings.LicenseType).ToLower() -ne "full"){
        $settings.LicenseType = Read-Host -Prompt "`rLicense type ('trial' or 'full')"
      }
    }
    else{
      Write-Host ""
    }
    
    if(($settings.LicenseType).ToLower() -eq "full"){
      $settings.SerialNumber = Read-Host -Prompt "`rSerial number"
    }

  }
  else{

    if(($settings.LicenseType).ToLower() -ne "trial" -And ($settings.LicenseType).ToLower() -ne "full"){
      $message = "Please specify a valid 'LicenseType' parameter. Valid values are 'trial' or 'full'."
      $message | Out-File -Append log.txt
      Write-Host $message
      exit 1
    }

  }
  $isIISInstalled = $true
  if((Get-WindowsOptionalFeature -Online | where FeatureName -Like  "IIS-WebServerRole").State -ne "Enabled"){
    $isIISInstalled = $false
    Enable-WindowsOptionalFeature -Online -FeatureName  "IIS-WebServerRole" | Out-File -Append log.txt
  }
  Import-module WebAdministration
  if($isIISInstalled -eq $false){
    Get-WebSite -Name "Default Web Site" | Remove-WebSite -Confirm:$false
    $message = "The Default Web Site has been removed in IIS."
    $message | Out-File -Append log.txt
  }
}

#Checking parameter validity
if (!$DownloadOnly -And $SkipParamInput -eq $true){
  if($settings.WebAppPath -eq $null -Or $settings.WebAppPath -eq ""){
    $message = "Please specify a 'WebAppPath' parameter."
    $message | Out-File -Append log.txt
    Write-Host $message
    exit 1
  }
  else{
    if (!(Test-Path($settings.WebAppPath))){
      try{
        New-Item $settings.WebAppPath -type directory -ErrorAction Stop | Out-File -Append log.txt
      }
      catch{
        $message = "Could not create a directory for parameter 'WebAppPath' at:" + $settings.WebAppPath
        $message | Out-File -Append log.txt
        Write-Host $message
        exit 1
      }
    }
  }
  if($settings.ServicesPath -eq $null -Or $settings.ServicesPath -eq ""){
    $message = "Please specify a 'ServicesPath' parameter."
    $message | Out-File -Append log.txt
    Write-Host $message
    exit 1
  }
  else{
    if (!(Test-Path($settings.ServicesPath))){
      try{
        New-Item $settings.ServicesPath -type directory -ErrorAction Stop | Out-File -Append log.txt
      }
      catch{
        $message = "Could not create a directory for parameter 'ServicesPath' at:" + $settings.ServicesPath
        $message | Out-File -Append log.txt
        Write-Host $message
        exit 1
      }
    }
  }
  if($settings.NodeJSPath -eq $null -Or $settings.NodeJSPath -eq ""){
    $message = "Please specify a 'NodeJSPath' parameter."
    $message | Out-File -Append log.txt
    Write-Host $message
    exit 1
  }
  else{
    if (!(Test-Path($settings.NodeJSPath))){
      try{
        New-Item $settings.NodeJSPath -type directory -ErrorAction Stop | Out-File -Append log.txt
      }
      catch{
        $message = "Could not create a directory for parameter 'NodeJSPath' at:" + $settings.NodeJSPath
        $message | Out-File -Append log.txt
        Write-Host $message
        exit 1
      }
    }
  }
  if($settings.SQLServerInstance -eq $null -Or $settings.SQLServerInstance -eq ""){
    $message = "Please specify a 'SQLServerInstance' parameter."
    $message | Out-File -Append log.txt
    Write-Host $message
    exit 1
  }
  if($settings.DBName -eq $null -Or $settings.DBName -eq ""){
    $message = "Please specify a 'DBName' parameter."
    $message | Out-File -Append log.txt
    Write-Host $message
    exit 1
  }
  if($settings.DBFilePath -ne $null -And $settings.DBFilePath -ne ""){
    if (!(Test-Path($settings.DBFilePath))){
      try{
        New-Item $settings.DBFilePath -type directory -ErrorAction Stop | Out-File -Append log.txt
      }
      catch{
        $message = "Could not create a directory for parameter 'DBFilePath' at:" + $settings.DBFilePath
        $message | Out-File -Append log.txt
        Write-Host $message
        exit 1
      }
    }
  }
  if($settings.DBLogPath -ne $null -And $settings.DBLogPath -ne ""){
    if (!(Test-Path($settings.DBLogPath))){
      try{
        New-Item $settings.DBLogPath -type directory -ErrorAction Stop | Out-File -Append log.txt
      }
      catch{
        $message = "Could not create a directory for parameter 'DBLogPath' at:" + $settings.DBLogPath
        $message | Out-File -Append log.txt
        Write-Host $message
        exit 1
      }
    }
  }
  if($settings.DBLoginID -eq $null -Or $settings.DBLoginID -eq ""){
    $message = "Please specify a 'DBLoginID' parameter."
    $message | Out-File -Append log.txt
    Write-Host $message
    exit 1
  }
  if($settings.DBPassword -eq $null -Or $settings.DBPassword -eq ""){
    $message = "Please specify a 'DBPassword' parameter."
    $message | Out-File -Append log.txt
    Write-Host $message
    exit 1
  }
  if($settings.SQLLoginID -ne $null -And $settings.SQLLoginID -ne ""){
    try{
        $sqlConnectionString = "Data Source=" + $settings.SQLServerInstance + ";User ID=" + $settings.SQLLoginID+ ";Password=" + $settings.SQLPassword + ";"
        $sqlConnection = new-object system.data.SqlClient.SQLConnection($sqlConnectionString)
        $sqlStateQuery = "SELECT count(*) FROM master.sys.databases WHERE name = '"+$settings.DBName+"'"
        $cmd = new-object system.data.sqlclient.sqlcommand($sqlStateQuery, $sqlConnection)
        $sqlConnection.Open()
        $testDbName = [convert]::ToInt32($cmd.ExecuteScalar(), 10)

        $message = "Connected to " + $settings.SQLServerInstance + " server succesfully."
        $message | Out-File -Append log.txt

        if ($testDbName -gt 0){
          $sqlConnection.Close()
          try{
            $message = "Database " + $settings.DBName + " exists."
            $message | Out-File -Append log.txt
            $sqlConnectionString = "Data Source=" + $settings.SQLServerInstance + ";Initial Catalog=" + $settings.DBName + ";User ID=" + $settings.DBLoginID+ ";Password=" + $settings.DBPassword + ";"
            $sqlConnection = new-object system.data.SqlClient.SQLConnection($sqlConnectionString)
            $sqlConnection.Open()
            $message = "Connected to " + $settings.DBName + " database succesfully via " + $settings.DBLoginID + "."
            $message | Out-File -Append log.txt
            if (CheckDbSchema){
              $message = $settings.DBName + " database schema is valid." | Out-File -Append log.txt
            }
            else{
              Write-Host $message
              $message = $settings.DBName + " database schema is invalid. Please ensure the database name is correct."
              $message | Out-File -Append log.txt
              Write-Host $message
              $message = "If the name is correct, there may have have been an issue creating the database during a previous installation. To re-use the same database name, please delete the database manually."
              $message | Out-File -Append log.txt
              Write-Host $message
              exit 1
            }
            $sqlConnection.Close()

          }
          catch{
            $message = "Could not connect to " + $settings.DBName + " database via " + $settings.DBLoginID + "."
            $message | Out-File -Append log.txt
            Write-Host $message
            exit 1
          }
        }
        else{
          $sqlCheckUser = "SELECT count(*) FROM sys.server_principals where name = '" + $settings.DBLoginID + "'"
          $cmd = new-object system.data.sqlclient.sqlcommand($sqlCheckUser, $sqlConnection)
          $testDbUser = [convert]::ToInt32($cmd.ExecuteScalar(), 10)
          $sqlConnection.Close()
          if ($testDbUser -gt 0){
            $message = "Database login " + $settings.DBLoginID + " already exists; a new login must be created for a new database."
            $message | Out-File -Append log.txt
            Write-Host $message
            exit 1
          }
        }
      }
      catch{
        $message = "Could not connect to " + $settings.SQLServerInstance + " via " + $settings.SQLLoginID + "."
        $message | Out-File -Append log.txt
        Write-Host $message
        exit 1
      }
  }
  else{
    try{
      $sqlConnectionString = "Data Source=" + $settings.SQLServerInstance + ";Initial Catalog=" + $settings.DBName + ";User ID=" + $settings.DBLoginID+ ";Password=" + $settings.DBPassword + ";"
      $sqlConnection = new-object system.data.SqlClient.SQLConnection($sqlConnectionString)
      $sqlConnection.Open()
      $message = "Connected to " + $settings.SQLServerInstance + " on " + $settings.DBName + " database succesfully via " + $settings.DBLoginID + "."
      $message | Out-File -Append log.txt
      if (CheckDbSchema){
        $message = $settings.DBName + " database schema is valid." | Out-File -Append log.txt
      }
      else{
        Write-Host $message
        $message = $settings.DBName + " database schema is invalid. Please ensure the database name is correct."
        $message | Out-File -Append log.txt
        Write-Host $message
        $message = "If the name is correct, there may have have been an issue creating the database during a previous installation. To re-use the same database name, please delete the database manually."
        $message | Out-File -Append log.txt
        Write-Host $message
        exit 1
      }
      $sqlConnection.Close()
    }
    catch{
      $message = "Could not connect to " + $settings.SQLServerInstance + " on database " + $settings.DBName + " via " + $settings.DBLoginID + "."
      $message | Out-File -Append log.txt
      Write-Host $message
    }
  }
  if($settings.WFGAdminUsername -eq $null -Or $settings.WFGAdminUsername -eq ""){
    $message = "Please specify a 'WFGAdminUsername' parameter."
    $message | Out-File -Append log.txt
    Write-Host $message
    exit 1
  }
  $filter = "LocalAccount='true' and Name='"+$settings.WFGAdminUsername+"'"
  if(($settings.IISAuthentication -eq "Basic" -Or  $settings.IISAuthentication -eq "Windows") -And ((Get-WmiObject Win32_UserAccount -Filter $filter) -eq $null) -And ($settings.WFGAdminPassword -eq $null -Or $settings.WFGAdminPassword -eq "")){
    $message = "Please specify a 'WFGAdminPassword' parameter."
    $message | Out-File -Append log.txt
    Write-Host $message
    exit 1
  }
  if($settings.IISSiteName -eq $null -Or $settings.IISSiteName -eq ""){
    $message = "Please specify a 'IISSiteName' parameter."
    $message | Out-File -Append log.txt
    Write-Host $message
    exit 1
  }
  if($settings.IISAppPoolName -eq $null -Or $settings.IISAppPoolName -eq ""){
    $message = "Please specify a 'IISAppPoolName' parameter."
    $message | Out-File -Append log.txt
    Write-Host $message
    exit 1
  }
  if($settings.IISBinding -eq $null -Or $settings.IISBinding -eq ""){
    $message = "Please specify a 'IISBinding' parameter."
    $message | Out-File -Append log.txt
    Write-Host $message
    exit 1
  }
  $WFGWebsite = (Get-WebSite $settings.IISSiteName)
  if($settings.IISPort -eq $null -Or $settings.IISPort -eq ""){
    $message = "Please specify a 'IISPort' parameter."
    $message | Out-File -Append log.txt
    Write-Host $message
    exit 1
  }
  else{
    $websites = Get-ChildItem  iis:\sites | Where {$_.State -eq "Started"}
    $binding = $settings.IISBinding + ":" + $settings.IISPort + ":"
    if ($WFGWebsite -ne $null -And $WFGWebsite.Bindings.Collection.bindingInformation -ne $binding){
      foreach ($site in $websites) {
        if([bool]($site.bindings.Collection | ? { $_.bindingInformation -eq "$binding" }) -eq $true){
          $message = "`rThe installation has been aborted due to the following error: The port " + $settings.IISPort + " is already bound to an existing website in IIS. Please specify another port in the configuration file."
          $message | Out-File -Append log.txt
          Write-Host $message
          exit 1
        }
      }
    }
  }

  if($settings.IISAuthentication -eq $null -Or $settings.IISAuthentication -eq "" -Or ($settings.IISAuthentication -ne "WorkflowGen" -And $settings.IISAuthentication -ne "Basic" -And $settings.IISAuthentication -ne "Windows")){
    $message = "The 'IISAuthentication' parameter is invalid. Valid values are : 'Windows', 'Basic' or 'WorkflowGen'."
    $message | Out-File -Append log.txt
    Write-Host $message
    exit 1
  }
  if ($WFGWebsite -ne $null -And $WFGWebsite.State -eq "Started"){
      $message = "The " + $settings.IISSiteName + " site is already running in IIS. It should be stopped before relaunching the installation."
      $message | Out-File -Append log.txt
      Write-Host $message
      exit 1
  }
}
elseif (!$DownloadOnly){
  Write-Host -NoNewLine "Web application path:" $settings.WebAppPath
  $key = $host.UI.RawUI.ReadKey("NoEcho,IncludeKeyDown")
  if ($key.VirtualKeyCode -ne 13){
    $settings.WebAppPath = ""
    while (($settings.WebAppPath).ToLower() -eq ""){
      Write-Host -NoNewLine "`r                                                                                                            "
      $settings.WebAppPath = Read-Host -Prompt "`rWeb application path (absolute path)"
      if (($settings.WebAppPath) -ne ""){
        if (!(Test-Path($settings.WebAppPath))){
          $create = Read-Host -Prompt "Path does not exist, would you like to create it? (Y/N)"
          if ($create.ToLower() -eq "y"){
            try{
              New-Item $settings.WebAppPath -type directory -ErrorAction Stop | Out-File -Append log.txt
            }
            catch{
              $message = "Could not create a directory at: " + $settings.WebAppPath
              $message | Out-File -Append log.txt
              Write-Host $message
              $settings.WebAppPath = ""
            }
          }
          else{
            $settings.WebAppPath = ""
          }
        }
      }
    }
  }
  else{
    Write-Host ""
  }
  Write-Host -NoNewLine "Windows services path:" $settings.ServicesPath
  $key = $host.UI.RawUI.ReadKey("NoEcho,IncludeKeyDown")
  if ($key.VirtualKeyCode -ne 13){
    $settings.ServicesPath = ""
    while (($settings.ServicesPath).ToLower() -eq ""){
      Write-Host -NoNewLine "`r                                                                                                            "
      $settings.ServicesPath = Read-Host -Prompt "`rWindows services path (absolute path)"
      if ($settings.ServicesPath -ne ""){
        if (!(Test-Path($settings.ServicesPath))){
          $create = Read-Host -Prompt "Path does not exist, would you like to create it? (Y/N)"
          if ($create.ToLower() -eq "y"){
            try{
              New-Item $settings.ServicesPath -type directory -ErrorAction Stop | Out-File -Append log.txt
            }
            catch{
              $message = "Could not create a directory at: " + $settings.ServicesPath
              $message | Out-File -Append log.txt
              Write-Host $message
              $settings.ServicesPath = ""
            }
          }
          else{
            $settings.ServicesPath = ""
          }
        }
      }
    }
  }
  else{
    Write-Host ""
  }
  Write-Host -NoNewLine "Nodejs path:" $settings.NodeJSPath
  $key = $host.UI.RawUI.ReadKey("NoEcho,IncludeKeyDown")
  if ($key.VirtualKeyCode -ne 13){
    $settings.NodeJSPath = ""
    while (($settings.NodeJSPath).ToLower() -eq ""){
      Write-Host -NoNewLine "`r                                                                                                            "
      $settings.NodeJSPath = Read-Host -Prompt "`rNodejs path (absolute path)"
      if ($settings.NodeJSPath -ne "")
      {
        if (!(Test-Path($settings.NodeJSPath))){
          $create = Read-Host -Prompt "Path does not exist, would you like to create it? (Y/N)"
          if ($create.ToLower() -eq "y"){
            try{
              New-Item $settings.NodeJSPath -type directory -ErrorAction Stop | Out-File -Append log.txt
            }
            catch{
              $message = "Could not create a directory at: " + $settings.NodeJSPath
              $message | Out-File -Append log.txt
              Write-Host $message
              $settings.NodeJSPath = ""
            }
          }
          else{
            $settings.NodeJSPath = ""
          }
        }
      }
    }
  }
  else{
    Write-Host ""
  }

  $dbServerValid = $false
  while ($dbServerValid -eq $false){
    Write-Host -NoNewLine "Database server instance:" $settings.SQLServerInstance
    $key = $host.UI.RawUI.ReadKey("NoEcho,IncludeKeyDown")
    if ($key.VirtualKeyCode -ne 13){
      $settings.SQLServerInstance = ""
      while ($settings.SQLServerInstance -eq ""){
        Write-Host -NoNewLine "`r                                                                                                            "
        $settings.SQLServerInstance = Read-Host -Prompt "`rDatabase server instance"
      }
    }
    else{
      Write-Host ""
    }
    Write-Host -NoNewLine "Database server login (required only if creating a new database):" $settings.SQLLoginID
    $key = $host.UI.RawUI.ReadKey("NoEcho,IncludeKeyDown")
    $changedSQLPassword = $false
    if ($key.VirtualKeyCode -ne 13){
      Write-Host -NoNewLine "`r                                                                                                            "
      $settings.SQLLoginID = Read-Host -Prompt "`rDatabase server login (required only if creating a new database)"
      $changedSQLPassword = $true
    }
    else{
      Write-Host ""
    }
    if ($settings.SQLLoginID -ne "" -And $settings.SQLLoginID -ne $null){
      if ($changedSQLPassword -Or $settings.SQLPassword -eq "" -Or $settings.SQLPassword -eq $null){
        $login = $settings.SQLLoginID
        $settings.SQLPassword = Read-Host -Prompt "Enter password for database server login '$login'" -AsSecureString
        $settings.SQLPassword = [Runtime.InteropServices.Marshal]::PtrToStringAuto([Runtime.InteropServices.Marshal]::SecureStringToBSTR($settings.SQLPassword))
      }
      try{
        $sqlConnectionString = "Data Source=" + $settings.SQLServerInstance + ";User ID=" + $settings.SQLLoginID + ";Password=" + $settings.SQLPassword + ";"
        $sqlConnection = new-object system.data.SqlClient.SQLConnection($sqlConnectionString)
        $sqlStateQuery = "SELECT count(*) FROM master.sys.databases WHERE name = '" + $settings.DBName + "'"
        $cmd = new-object system.data.sqlclient.sqlcommand($sqlStateQuery, $sqlConnection)
        $sqlConnection.Open()
        $testDbName = [convert]::ToInt32($cmd.ExecuteScalar(), 10)
        $sqlConnection.Close()
        $message = "Connected to " + $settings.SQLServerInstance + " server succesfully."
        $message | Out-File -Append log.txt
        $dbServerValid = $true
      }
      catch{
        $message = "Could not connect to " + $settings.SQLServerInstance + " via " + $settings.SQLLoginID + "."
        $message | Out-File -Append log.txt
        Write-Host $message
      }
    }
    else{
      $dbServerValid = $null
    }
  }

  $dbValid = $false
  while ($dbValid -eq $false){
    Write-Host -NoNewLine "Database name:" $settings.DBName
    $key = $host.UI.RawUI.ReadKey("NoEcho,IncludeKeyDown")
    if ($key.VirtualKeyCode -ne 13){
      $settings.DBName = ""
      while ($settings.DBName -eq ""){
        Write-Host -NoNewLine "`r                                                                                                            "
        $settings.DBName = Read-Host -Prompt "`rDatabase name"
        $testDbName = 0
        if ($dbServerValid -eq $true){
          $sqlConnectionString = "Data Source=" + $settings.SQLServerInstance + ";User ID=" + $settings.SQLLoginID + ";Password=" + $settings.SQLPassword + ";"
          $sqlConnection = new-object system.data.SqlClient.SQLConnection($sqlConnectionString)
          $sqlConnection.Open()

          $sqlStateQuery = "SELECT count(*) FROM master.sys.databases WHERE name = '" + $settings.DBName + "'"
          $cmd = new-object system.data.sqlclient.sqlcommand($sqlStateQuery, $sqlConnection)
          $testDbName = [convert]::ToInt32($cmd.ExecuteScalar(), 10)
          $sqlConnection.Close()
        }
      }
    }
    else{
      Write-Host ""
    }

    Write-Host -NoNewLine "Database user login:" $settings.DBLoginID
    $key = $host.UI.RawUI.ReadKey("NoEcho,IncludeKeyDown")
    $changedDBPassword = $false
    if ($key.VirtualKeyCode -ne 13 -Or $settings.DBLoginID -eq $null -Or $settings.DBLoginID -eq ""){
      $settings.DBLoginID = ""
      while ($settings.DBLoginID -eq ""){
        Write-Host -NoNewLine "`r                                                                                                            "
        $settings.DBLoginID = Read-Host -Prompt "`rDatabase user login"
        $changedDBPassword = $true
      }
    }
    else{
      Write-Host ""
    }
    if ($settings.DBLoginID -ne "" -And $settings.DBLoginID -ne $null){
      if($changedDBPassword -Or $newDBUser -eq $true -Or $settings.DBPassword -eq $null -Or $settings.DBPassword -eq ""){
        $login = $settings.DBLoginID
        $settings.DBPassword = Read-Host -Prompt "Enter password for database user login '$login'" -AsSecureString
        $settings.DBPassword = [Runtime.InteropServices.Marshal]::PtrToStringAuto([Runtime.InteropServices.Marshal]::SecureStringToBSTR($settings.DBPassword))
      }
    }

    if ($dbServerValid -eq $true){
      if ($testDbName -gt 0){
        try{
          $message = "Database " + $settings.DBName + " exists."
          $message | Out-File -Append log.txt
          $sqlConnectionString = "Data Source=" + $settings.SQLServerInstance + ";Initial Catalog=" + $settings.DBName + ";User ID=" + $settings.DBLoginID+ ";Password=" + $settings.DBPassword + ";"
          $sqlConnection = new-object system.data.SqlClient.SQLConnection($sqlConnectionString)
          $sqlConnection.Open()
          $message = "Connected to " + $settings.DBName + " database succesfully via " + $settings.DBLoginID + "."
          $message | Out-File -Append log.txt
          if (CheckDbSchema){
            $message = $settings.DBName + " database schema is valid." | Out-File -Append log.txt
            $dbValid = $true
          }
          else{
            Write-Host $message
            $message = $settings.DBName + " database schema is invalid. Please ensure the database name is correct."
            $message | Out-File -Append log.txt
            Write-Host $message
            $message = "If the name is correct, there may have have been an issue creating the database during a previous installation. To re-use the same database name, please delete the database manually."
            $message | Out-File -Append log.txt
            Write-Host $message
            $dbValid = $false
          }
          $sqlConnection.Close()
        }
        catch{
          $message = "Could not connect to " + $settings.DBName + " database via " + $settings.DBLoginID + "."
          $message | Out-File -Append log.txt
          Write-Host $message
        }
      }
      else{
        try{
          $sqlConnectionString = "Data Source=" + $settings.SQLServerInstance + ";User ID=" + $settings.SQLLoginID + ";Password=" + $settings.SQLPassword + ";"
          $sqlConnection = new-object system.data.SqlClient.SQLConnection($sqlConnectionString)
          $sqlCheckUser = "SELECT count(*) FROM sys.server_principals where name = '" + $settings.DBLoginID + "'"
          $sqlConnection.Open()
          $cmd = new-object system.data.sqlclient.sqlcommand($sqlCheckUser, $sqlConnection)
          $testDbUser = [convert]::ToInt32($cmd.ExecuteScalar(), 10)
          $sqlConnection.Close()
          if ($testDbUser -gt 0){
            $message = "Database login " + $settings.DBLoginID + " already exists; a new login must be created for a new database."
            $message | Out-File -Append log.txt
            Write-Host $message
          }
          else{
            $newDBUser = $true
            if($changedDBPassword -Or $newDBUser -eq $true -Or $settings.DBPassword -eq $null -Or $settings.DBPassword -eq ""){
              $login = $settings.DBLoginID
              $DBPasswordVerify = Read-Host -Prompt "Confirm password for database user login '$login'" -AsSecureString
              $DBPasswordVerify = [Runtime.InteropServices.Marshal]::PtrToStringAuto([Runtime.InteropServices.Marshal]::SecureStringToBSTR($DBPasswordVerify))
              if ($DBPasswordVerify -ne $settings.DBPassword){
                $message = "Passwords do not match. Please verify the database information."
                $message | Out-File -Append log.txt
                Write-Host $message
                $dbValid = $false;
              }
              else{
                $dbValid = $true
              }
            }
          }
        }
        catch{
        }
      }
    }
    else{
      try{
        $sqlConnectionString = "Data Source=" + $settings.SQLServerInstance + ";Initial Catalog=" + $settings.DBName + ";User ID=" + $settings.DBLoginID + ";Password=" + $settings.DBPassword + ";"
        $sqlConnection = new-object system.data.SqlClient.SQLConnection($sqlConnectionString)
        $sqlConnection.Open()
        $message = "Connected to " + $settings.SQLServerInstance + " on " + $settings.DBName + " database succesfully via " + $settings.DBLoginID + "."
        $message | Out-File -Append log.txt
        if (CheckDbSchema){
          $message = $settings.DBName + " database schema is valid." | Out-File -Append log.txt
          $testDbName = 1
          $dbValid = $true
        }
        else{
          Write-Host $message
          $message = $settings.DBName + " database schema is invalid. Please ensure the database name is correct."
          $message | Out-File -Append log.txt
          Write-Host $message
          $message = "If the name is correct, there may have have been an issue creating the database during a previous installation. To re-use the same database name, please delete the database manually."
          $message | Out-File -Append log.txt
          Write-Host $message
          $dbValid = $false
        }
        $sqlConnection.Close()
      }
      catch{
        $message = "Could not connect to " + $settings.SQLServerInstance + " on database " + $settings.DBName + " via " + $settings.DBLoginID + "."
        $message | Out-File -Append log.txt
        Write-Host $message
      }
    }
  }

  if ($testDbName -lt 1){
    Write-Host -NoNewLine "Database file path (database instance default will be used if left empty):" $settings.DBFilePath
    $key = $host.UI.RawUI.ReadKey("NoEcho,IncludeKeyDown")
    if ($key.VirtualKeyCode -ne 13){
      $continue = $false
      while (!($continue)){
        Write-Host -NoNewLine "`r                                                                                                            "
        $settings.DBFilePath = Read-Host -Prompt "`rDatabase file path (database instance default will be used if left empty)"
        if ($settings.DBFilePath -ne "" -And !(Test-Path($settings.DBFilePath))){
          $create = Read-Host -Prompt "Path does not exist, would you like to create it? (Y/N)"
          if ($create.ToLower() -eq "y"){
            try{
              New-Item $settings.DBFilePath -type directory -ErrorAction Stop | Out-File -Append log.txt
              $continue = $true
            }
            catch{
              $message = "Could not create a directory at: " + $settings.DBFilePath
              $message | Out-File -Append log.txt
              Write-Host $message
              $continue = $false
            }
          }
          else{
            $continue = $false
          }
        }
        else{
          $continue = $true
        }
      }
    }
    else{
      Write-Host ""
    }
    Write-Host -NoNewLine "Database log path (database instance default will be used if left empty):" $settings.DBLogPath
    $key = $host.UI.RawUI.ReadKey("NoEcho,IncludeKeyDown")
    if ($key.VirtualKeyCode -ne 13){
      $continue = $false
      while (!($continue)){
        Write-Host -NoNewLine "`r                                                                                                            "
        $settings.DBLogPath = Read-Host -Prompt "`rDatabase log path (database instance default will be used if left empty)"
        if ($settings.DBLogPath -ne "" -And !(Test-Path($settings.DBLogPath))){
          $create = Read-Host -Prompt "Path does not exist, would you like to create it? (Y/N)"
          if ($create.ToLower() -eq "y"){
            try{
              New-Item $settings.DBLogPath -type directory -ErrorAction Stop | Out-File -Append log.txt
              $continue = $true
            }
            catch{
              $message = "Could not create a directory at: " + $settings.DBLogPath
              $message | Out-File -Append log.txt
              Write-Host $message
              $continue = $false
            }
          }
          else{
            $continue = $false
          }
        }
        else{
          $continue = $true
        }
      }
    }
    else{
      Write-Host ""
    }
  }
  $isExit = $false
  Do{
    Write-Host -NoNewLine "IIS site name:" $settings.IISSiteName
    $key = $host.UI.RawUI.ReadKey("NoEcho,IncludeKeyDown")
    if ($key.VirtualKeyCode -ne 13){
      $settings.IISSiteName = ""
      while ($settings.IISSiteName -eq ""){
        Write-Host -NoNewLine "`r                                                                                                            "
        $settings.IISSiteName = Read-Host -Prompt "`rIIS site name"
      }
      $isExit = $true
    }
    else{
      Write-Host ""
      $isExit = $true
    }
    $WFGWebsite = (Get-WebSite $settings.IISSiteName)
    if ($WFGWebsite -ne $null -And $WFGWebsite.State -eq "Started"){
      $siteName = $settings.IISSiteName
      $stopSite = ""
      while ($stopSite.ToLower() -ne "y" -And $stopSite.ToLower() -ne "n"){
        $stopSite = Read-Host -Prompt "`rThe $siteName site is already running in IIS. Would you like to stop it? (Y/N)"
      }
      if($stopSite.ToLower() -eq "y"){
        $message = "Stopping site " + $settings.IISSiteName
        Stop-WebSite $settings.IISSiteName
        Write-Host "`r$message. Done."
        $message = "The $siteName site has been stopped in IIS."
        $message | Out-File -Append log.txt
        $isExit = $true
      }
      else{
        $message = "Please enter another name for the website."
        Write-Host $message | Out-File -Append log.txt
        $isExit = $false
      }

    }
  } while(!$isExit)


  Write-Host -NoNewLine "IIS application pool name:" $settings.IISAppPoolName
  $key = $host.UI.RawUI.ReadKey("NoEcho,IncludeKeyDown")
  if ($key.VirtualKeyCode -ne 13){
    $settings.IISAppPoolName = ""
    while ($settings.IISAppPoolName -eq ""){
      Write-Host -NoNewLine "`r                                                                                                            "
      $settings.IISAppPoolName = Read-Host -Prompt "`rIIS application pool name"
    }
  }
  else{
    Write-Host ""
  }
  Write-Host -NoNewLine "IIS binding:" $settings.IISBinding
  $key = $host.UI.RawUI.ReadKey("NoEcho,IncludeKeyDown")
  if ($key.VirtualKeyCode -ne 13){
    $settings.IISBinding = ""
    while ($settings.IISBinding -eq ""){
      Write-Host -NoNewLine "`r                                                                                                            "
      $settings.IISBinding = Read-Host -Prompt "`rIIS binding"
    }
  }
  else{
    Write-Host ""
  }
  Write-Host -NoNewLine "IIS port:" $settings.IISPort
  $key = $host.UI.RawUI.ReadKey("NoEcho,IncludeKeyDown")
  if ($key.VirtualKeyCode -ne 13){
    $settings.IISPort = ""
    while ($settings.IISPort -eq ""){
      Write-Host -NoNewLine "`r                                                                                                            "
      $settings.IISPort = Read-Host -Prompt "`rIIS port"
    }
  }
  else{
    Write-Host ""
  }

  $isUsed = $false
  $websites =  Get-ChildItem  iis:\sites | Where {$_.State -eq "Started"}

  Do{
    $binding = $settings.IISBinding + ":" + $settings.IISPort + ":"
    $Port = $settings.IISPort
    foreach ($Site in $websites) {
      if([bool]($site.bindings.Collection | ? { $_.bindingInformation -eq "$binding" }) -eq $true){
        Do{
          Write-Host -NoNewLine "`r                                                                                                            "
          $settings.IISPort = Read-Host -Prompt "`rThe port $Port is already binded to an existing website in IIS. Please specify another port"
          $isUsed = $true
        }while ($settings.IISPort -eq "")
        if($isUsed){
          break
        }
      }
      else{
        $message = "Port $Port is operational"
        $message | Out-File -Append log.txt
        $isUsed = $false
      }
    }
  }while ($isUsed -eq $true)

  Write-Host -NoNewLine "IIS authentication:" $settings.IISAuthentication
  $key = $host.UI.RawUI.ReadKey("NoEcho,IncludeKeyDown")
  if ($key.VirtualKeyCode -ne 13){
    $settings.IISAuthentication = ""
    while ($settings.IISAuthentication -ne "windows" -And $settings.IISAuthentication -ne "basic" -And $settings.IISAuthentication -ne "workflowgen"){
      Write-Host -NoNewLine "`r                                                                                                   "
      $settings.IISAuthentication = (Read-Host -Prompt "`rIIS authentication ('Windows','Basic' or 'WorkflowGen')").ToLower()
    }
  }
  else{
    Write-Host ""
  }
  if ($settings.IISAuthentication -ne "workflowgen"){
    Write-Host -NoNewLine "WorkflowGen administrator domain (optional):" $settings.WFGAdminDomain
    $key = $host.UI.RawUI.ReadKey("NoEcho,IncludeKeyDown")
    if ($key.VirtualKeyCode -ne 13){
      $settings.WFGAdminDomain = ""
      Write-Host -NoNewLine "`r                                                                                            "
      $settings.WFGAdminDomain = Read-Host -Prompt "`rWorkflowGen user domain (not required)"
    }
    else{
      Write-Host ""
    }
  }
  Write-Host -NoNewLine "WorkflowGen administrator login:" $settings.WFGAdminUsername
  $key = $host.UI.RawUI.ReadKey("NoEcho,IncludeKeyDown")
  $changedWFGPassword = $false
  if ($key.VirtualKeyCode -ne 13){
    Write-Host -NoNewLine "`r                                                                                            "
    $settings.WFGAdminUsername = ""
    while($settings.WFGAdminUsername -eq ""){
      $settings.WFGAdminUsername = Read-Host -Prompt "`rWorkflowGen user login"
      $changedWFGPassword = $true
    }
  }
  else{
    Write-Host ""
  }
  $filter = "LocalAccount='true' and Name='"+$settings.WFGAdminUsername+"'"
  if (($settings.IISAuthentication -eq "Basic" -Or $settings.IISAuthentication -eq "Windows") -And ((Get-WmiObject Win32_UserAccount -Filter $filter) -eq $null)  -And ($changedWFGPassword -Or $settings.WFGAdminPassword -eq "" -Or $settings.WFGAdminPassword -eq $null)){
    $login = $settings.WFGAdminUsername
    Do{
      $settings.WFGAdminPassword = Read-Host -Prompt "Enter password for WorkflowGen administrator login '$login'" -AsSecureString
      $settings.WFGAdminPassword = [Runtime.InteropServices.Marshal]::PtrToStringAuto([Runtime.InteropServices.Marshal]::SecureStringToBSTR($settings.WFGAdminPassword))

      $DBPasswordVerify = Read-Host -Prompt "Confirm password for WorkflowGen administrator login '$login'" -AsSecureString
      $DBPasswordVerify = [Runtime.InteropServices.Marshal]::PtrToStringAuto([Runtime.InteropServices.Marshal]::SecureStringToBSTR($DBPasswordVerify))

      if ($DBPasswordVerify -ne $settings.WFGAdminPassword){
        $message = "Passwords do not match."
        $message | Out-File -Append log.txt
        Write-Host $message
        $settings.WFGAdminPassword = ""
      }
    } while($settings.WFGAdminPassword -eq "")
  }

  $overwrite = ""
  Write-Host "`n"
  while ($overwrite.ToLower() -ne "y" -And $overwrite.ToLower() -ne "n"){
    $overwrite = Read-Host -Prompt "Would you like to overwrite the current configuration file with the new values (passwords will be saved in plain text)? (Y/N)"
  }
  Write-Host "`n"
  if ($overwrite.ToLower() -eq "y"){
    try{
      $settings | ConvertTo-Json | Out-File config.json
    }
    catch{
      $settings | ConvertTo-Json -Compress | Out-File config.json #WS2012 workaround
    }
  }
  else{
    try{
      $settings | ConvertTo-Json | Out-File tempconfig.json
    }
    catch{
      $settings | ConvertTo-Json -Compress | Out-File tempconfig.json
    }
  }
}

# Detect SQL Server PowerShell module
Write-Host ""
$message = "Detecting required SQL Server PowerShell module..."
$message | Out-File -Append log.txt
Write-Host $message
if ($PSVersionTable.PSVersion.Major -ge 5) {
  try {
    if (Get-Module SqlServer -ListAvailable) {
      $message = "The SQL Server PowerShell module is already installed."
      $message | Out-File -Append log.txt
      Write-Host $message
    } 
    else {
      $message = "Installing SQL Server PowerShell module..."
      $message | Out-File -Append log.txt
      Write-Host $message

      # PowerShellGet requires NuGet provider version '2.8.5.201' or newer to interact with NuGet-based repositories. 
      Install-PackageProvider -Name NuGet -MinimumVersion 2.8.5.201 -Force

      # Trust the PSGallery repository
      Set-PSRepository -Name PSGallery -InstallationPolicy Trusted

      Install-Module -Name SqlServer
    }
    Import-Module SqlServer
  }
  catch {
    $message = "An error occurred while referencing the SQL Server PowerShell module."
    $message | Out-File -Append log.txt
    Write-Host $message "See log.txt for more details."
    $message = "ERROR: " + $PSItem.Exception.Message
    $message | Out-File -Append log.txt
  }
} else {
  try {
    Invoke-Sqlcmd
    $message = "The SQL Server PowerShell module is already installed."
    $message | Out-File -Append log.txt
    Write-Host $message
  }
  catch {
    $message = "The 'Invoke-Sqlcmd' cmdlet required by the installation script for the database creation was not found."
    $message | Out-File -Append log.txt
    Write-Host $message "See log.txt for more details."
    $message = "ERROR: " + $PSItem.Exception.Message
    $message | Out-File -Append log.txt
    Write-Host ""
    $message = "Please refer to the 'Prerequisites' section in the README.md guide to install the required SQL Server PowerShell module."
    $message | Out-File -Append log.txt
    Write-Host $message
    $message = "Once the SQL Server PowerShell module is installed, open a new PowerShell command window to relaunch the installation script."
    $message | Out-File -Append log.txt
    Write-Host $message
    exit 1
  }
}
Write-Host ""

try{
  if ((Get-WmiObject -Class Win32_Service -Filter "Name='w3svc'") -ne $null){
    iisreset | Out-File -Append log.txt
  }

  #Support different TLS versions (for https download urls)
  [Net.ServicePointManager]::SecurityProtocol = "tls12, tls11, tls"

  #Check if already downloaded
  if(!(Test-Path -Path "$package\manual.zip")){
    $message = "1.  Downloading WorkflowGen installation pack..."
    $message | Out-File -Append log.txt
    Write-Host -NoNewline $message
    Invoke-WebRequest $($constants.WFGManualZipPath) -OutFile "$package\manual.zip" -UseBasicParsing
    Write-Host "`r$message Done."
  }
  else{
    $message = "1.  WorkflowGen installation already downloaded."
    $message | Out-File -Append log.txt
    Write-Host $message
  }
  if(!(Test-Path -Path "$package\WFG-TRIAL.lic") -And $settings.LicenseType -eq "trial" ){
    $message = "2.  Downloading WorkflowGen Trial license file..."
    $message | Out-File -Append log.txt
    Write-Host -NoNewline $message
    Invoke-WebRequest $($constants.WFGTrialLicensePath) -OutFile "$package\WFG-TRIAL.lic" -UseBasicParsing
    Write-Host "`r$message Done."
  }
  elseif ($settings.LicenseType -eq "trial" ){
    $message = "2.  WorkflowGen Trial license file already downloaded."
    $message | Out-File -Append log.txt
    Write-Host $message
  }
  if(!(Test-Path -Path "$package\iisnode.msi")){
    $message = "3.  Downloading IISNode installer..."
    $message | Out-File -Append log.txt
    Write-Host -NoNewline $message
    Invoke-WebRequest $($constants.IISNodeURL) -OutFile "$package\iisnode.msi" -UseBasicParsing
    Write-Host "`r$message Done."
  }
  else{
    $message = "3.  IISNode installer already downloaded."
    $message | Out-File -Append log.txt
    Write-Host $message
  }

  if(!(Test-Path -Path "$package\rewrite.msi")){
    $message = "4.  Downloading UrlRewrite (IIS module) installer..."
    $message | Out-File -Append log.txt
    Write-Host -NoNewline $message
    Invoke-WebRequest $($constants.URLRewriteURL) -OutFile "$package\rewrite.msi" -UseBasicParsing
    Write-Host "`r$message Done."
  }
  else{
    $message = "4.  UrlRewrite (IIS module) installer already downloaded."
    $message | Out-File -Append log.txt
    Write-Host $message
  }

  if (!(Test-Path "$package\node.zip" )){
    $message = "5.  Downloading Node.js..."
    $message | Out-File -Append log.txt
    Write-Host -NoNewline $message
    Invoke-WebRequest $($constants.NodeJSURL -f $constants.NodeJSVersion) -OutFile "$package\node.zip" -UseBasicParsing
    Write-Host "`r$message Done."
  }
  else{
    $message = "5.  Node.js is already downloaded."
    $message | Out-File -Append log.txt
    Write-Host $message
  }

  if ($DownloadOnly){
    $message = "Download complete."
    $message | Out-File -Append log.txt
    Write-Host "`n"
    Write-Host $message
    break;
  }

  #Check if already extrated
  if(!(Test-Path -Path "$package\manual")){
    $message = "6.  Extracting WorkflowGen installation pack..."
    $message | Out-File -Append log.txt
    Write-Host -NoNewline $message
    (New-object -com shell.application).namespace($package).CopyHere((New-object -com shell.application).namespace("$package\manual.zip").Items(),20);
    Write-Host "`r$message Done."
  }
  else{
    $message = "6.  WorkflowGen installation pack already extracted."
    $message | Out-File -Append log.txt
    Write-Host $message
  }

  # Unblock the script files
  Unblock-File -Path ".\scripts\build.ps1"
  Unblock-File -Path ".\scripts\start.ps1"

  Invoke-Expression ".\scripts\build.ps1"
  Invoke-Expression ".\scripts\start.ps1"
}
catch{
  Write-Host "`nAn error occurred, details are located in the log file."
  Write-Host $_
  $_ | Out-File -Append log.txt
  exit $LastExitCode
}
