if (Test-Path ".\tempconfig.json"){
  $Docker = $false
  $settings = (Get-Content ".\tempconfig.json" -Raw) | ConvertFrom-Json
  $settings.IISAuthentication = ($settings.IISAuthentication).ToLower()
}
elseif (Test-Path ".\config.json"){
  $Docker = $false
  $settings = (Get-Content ".\config.json" -Raw) | ConvertFrom-Json
  $settings.IISAuthentication = ($settings.IISAuthentication).ToLower()
}
else{
  $Docker = $true
  $settings.WFGWebAppPath = "C:\Inetpub\wwwroot\wfgen"
}
$ErrorActionPreference = "Stop";

$pwd = (Get-Item -Path ".\" -Verbose).FullName
$package = $pwd + "\package"

$SQLInputFilePath = "$package/manual/Databases/MsSQLServer/Create_WFG-V7-0_SQL_"
$SQLSuffixTables = 'Tables.sql'
$SQLSuffixPKeys = 'PKeys.sql'
$SQLSuffixFKeys = 'FKeys.sql'
$SQLSuffixIndexes = 'Indexes.sql'
$SQLSuffixTriggers = 'Triggers.sql'
$SQLSuffixConst = 'Const.sql'

# start the service
if ($Docker -eq $true){
	Write-Verbose "Starting SQL Server"
	start-service MSSQL`$SQLEXPRESS
}

$message = "16. Installing IISNode..."
$message | Out-File -Append log.txt
Write-Host -NoNewLine $message
Start-Process -FilePath "$package\iisnode.msi" /qn -Wait
Write-Host "`r$message Done."

$message = "17. Installing URLRewrite for IIS..."
$message | Out-File -Append log.txt
Write-Host -NoNewLine $message
Start-Process -FilePath "$package\rewrite.msi" /qn -Wait
Write-Host "`r$message Done."

if ($Docker -eq $true){
	Write-Verbose "Changing SA login credentials"
	$sqlcmd = "ALTER LOGIN sa WITH password='" + $settings.SQLPassword + "';ALTER LOGIN sa ENABLE;"
	Invoke-Sqlcmd -Query $sqlcmd -ServerInstance $settings.SQLServerInstance
}

$DBStateQuery = "SELECT name FROM master.sys.databases WHERE name = '"+$settings.DBName+"'"
$DBCreateQuery = "CREATE DATABASE "+$settings.DBName
if ($settings.DBFilePath -ne $null -And $settings.DBFilePath -ne ""){
	$DBCreateQuery += "	ON (NAME = "+$settings.DBName+"_dat, FILENAME = '"+$settings.DBFilePath+"\"+$settings.DBName+".mdf')"
}
if($settings.DBLogPath -ne $null -And $settings.DBLogPath -ne ""){
	$DBCreateQuery += "	LOG ON (NAME = "+$settings.DBName+"_log, FILENAME = '"+$setings.DBLogPath+"\"+$settings.DBName+"_log.ldf')"
}
$DBUserCreate1 = "CREATE LOGIN "+$settings.DBLoginID+" WITH PASSWORD = '"+$settings.DBPassword+"';"
$DBUserCreate2 = "CREATE USER "+$settings.DBLoginID+" FOR LOGIN "+$settings.DBLoginID+";"
$DBUserCreate3 = "exec sp_addrolemember N'db_datareader', "+$settings.DBLoginID+"; exec sp_addrolemember N'db_datawriter', "+$settings.DBLoginID+";"

if($settings.SQLLoginID -ne $null -And $settings.SQLLoginID -ne ""){
  push-location
  $DBState = Invoke-Sqlcmd -ServerInstance $settings.SQLServerInstance -Username $settings.SQLLoginID -Password $settings.SQLPassword -Query $DBStateQuery
  pop-location
  if ($DBState -eq $null){
    $message = "18. Creating database: " + $settings.DBName + "..."
    $message | Out-File -Append log.txt
    Write-Host -NoNewLine $message
    Invoke-Sqlcmd -ServerInstance $settings.SQLServerInstance -Username $settings.SQLLoginID -Password $settings.SQLPassword -Query $DBCreateQuery
    Write-Host "`r$message Done."

    $message = "19. Creating database user: " + $settings.DBLoginID + "..."
    $message | Out-File -Append log.txt
    Write-Host -NoNewLine $message
    # Attempting to create the database user using Invoke-Sqlcmd would sometimes fail on remote sql connections (invalid login).
    # Using sqlclient is more reliable, however this method cannot be used for runnning the database creation scripts that contain "GO" statements, which is not TSQL.
    $sqlConnectionString = "Data Source=" + $settings.SQLServerInstance + ";Initial Catalog=" + $settings.DBName + ";User ID=" + $settings.SQLLoginID+ ";Password=" + $settings.SQLPassword + ";"
    $sqlConnection = new-object system.data.SqlClient.SQLConnection($sqlConnectionString);
    $cmd = new-object system.data.sqlclient.sqlcommand($DBStateQuery, $sqlConnection);
    $sqlConnection.Open()
    (new-object system.data.sqlclient.sqlcommand($DBUserCreate1, $sqlConnection)).ExecuteNonQuery() > $null
    (new-object system.data.sqlclient.sqlcommand($DBUserCreate2, $sqlConnection)).ExecuteNonQuery() > $null
    (new-object system.data.sqlclient.sqlcommand($DBUserCreate3, $sqlConnection)).ExecuteNonQuery() > $null
    Write-Host "`r$message Done."

    $sqlConnection.Close()
    $message = "20. Creating database tables..."
    $message | Out-File -Append log.txt
    Write-Host -NoNewLine $message
  	Invoke-Sqlcmd -ServerInstance $settings.SQLServerInstance -Username $settings.SQLLoginID -Password $settings.SQLPassword -InputFile $SQLInputFilePath$SQLSuffixTables -Database $settings.DBName
    Write-Host "`r$message Done."

    $message = "21. Creating database primary keys..."
    $message | Out-File -Append log.txt
    Write-Host -NoNewLine $message
  	Invoke-Sqlcmd -ServerInstance $settings.SQLServerInstance -Username $settings.SQLLoginID -Password $settings.SQLPassword  -InputFile $SQLInputFilePath$SQLSuffixPKeys -Database $settings.DBName
    Write-Host "`r$message Done."

    $message = "22. Creating database foreign keys..."
    $message | Out-File -Append log.txt
    Write-Host -NoNewLine $message
  	Invoke-Sqlcmd -ServerInstance $settings.SQLServerInstance -Username $settings.SQLLoginID -Password $settings.SQLPassword -InputFile $SQLInputFilePath$SQLSuffixFKeys -Database $settings.DBName
    Write-Host "`r$message Done."

    $message = "23. Creating database indexes..."
    $message | Out-File -Append log.txt
    Write-Host -NoNewLine $message
  	Invoke-Sqlcmd -ServerInstance $settings.SQLServerInstance -Username $settings.SQLLoginID -Password $settings.SQLPassword -InputFile $SQLInputFilePath$SQLSuffixIndexes -Database $settings.DBName
    Write-Host "`r$message Done."

    $message = "24. Creating database triggers..."
    $message | Out-File -Append log.txt
    Write-Host -NoNewLine $message
  	Invoke-Sqlcmd -ServerInstance $settings.SQLServerInstance -Username $settings.SQLLoginID -Password $settings.SQLPassword -InputFile $SQLInputFilePath$SQLSuffixTriggers -Database $settings.DBName
    Write-Host "`r$message Done."

    $message = "25. Creating database constants..."
    $message | Out-File -Append log.txt
    Write-Host -NoNewLine $message
  	Invoke-Sqlcmd -ServerInstance $settings.SQLServerInstance -Username $settings.SQLLoginID -Password $settings.SQLPassword -InputFile $SQLInputFilePath$SQLSuffixConst -Database $settings.DBName
    Write-Host "`r$message Done."
  }
  else{
    $message = "18. Database already exists: " + $settings.DBName
    $message | Out-File -Append log.txt
    Write-Host $message
  }
}
else{
  $message = "18. No Database server credentials provided. Cannot confirm that database exists."
  $message | Out-File -Append log.txt
  Write-Host $message
}

if ($Docker -eq $true -Or $settings.IISAuthentication -eq "workflowgen"){
  Invoke-Sqlcmd -ServerInstance $settings.SQLServerInstance -Username $settings.DBLoginID -Password $settings.DBPassword -Query "UPDATE USERS SET password = '7fa48cdd53abafc8c96cb38b11052bff9da56142d484270ccad9d92fd234515e', salt = 's/edDsXtiv9IcEph9E7IJHPbUcY+/uOSerxRPFQEDdA=' WHERE ID_USER = 1; UPDATE DIRECTORY SET AUTH ='Y' WHERE DIRNAME='WORKFLOWGEN';" -Database $settings.DBName
}
else{
  Invoke-Sqlcmd -ServerInstance $settings.SQLServerInstance -Username $settings.DBLoginID -Password $settings.DBPassword -Query ("UPDATE USERS SET username ='"+$settings.WFGAdminUsername+"' WHERE ID_USER = 1;") -Database $settings.DBName
}

$WFGConnectionString = "Data Source=" + $settings.SQLServerInstance + ";Initial Catalog=" + $settings.DBName + ";User ID=" + $settings.DBLoginID + ";Password=" + $settings.DBPassword + ";"

#Modify Web.Config Connection String
$message = "26. Modifying the wfgen web.config..."
$message | Out-File -Append log.txt
Write-Host -NoNewLine $message
$webConfig = $settings.WebAppPath + "\web.config"
$appConfig = [xml](cat $webConfig)
$appConfig.configuration.connectionStrings.add | foreach {
    $_.connectionString = $WFGConnectionString
}


if($settings.IISPort -eq "80" -Or $settings.IISPort -contains ":"){
  $websiteUrl = $settings.WebsiteURL
}
else{
  $uri = ([uri]$settings.WebsiteURL)
  $websiteUrl = $uri.Scheme + "://" + $uri.Host + ":" + $settings.IISPort + $uri.AbsolutePath
}

#Change Web.Config Application Settings
$foundSettings = $appConfig.configuration.appSettings.add | ? { $_.key -eq "ApplicationUrl" }
if ($foundSettings.Count -ne 0) {
  $foundSettings | % { $_.value = $websiteUrl }
}

if ($settings.SerialNumber -ne $null){
	$foundSettings = $appConfig.configuration.appSettings.add | ? { $_.key -eq "ApplicationSerialNumber"}
	if ($foundSettings.Count -ne 0) {
	  $foundSettings | % { $_.value = $settings.SerialNumber }
	}
}

$foundSettings = $appConfig.configuration.appSettings.add | ? { $_.key -eq "ApplicationSmtpServer"}
if ($foundSettings.Count -ne 0) {
  $foundSettings | % { $_.value = $settings.SMTPServer }
}

$foundSettings = $appConfig.configuration.appSettings.add | ? { $_.key -eq "EngineNotificationDefaultSender"}
if ($foundSettings.Count -ne 0) {
  $foundSettings | % { $_.value = $settings.DefaultEmailSender }
}

if ($settings.IISAuthentication -eq "workflowgen" -And $appConfig.configuration.SelectNodes('system.webServer').Count -eq 0){
  $webServerNode = $appConfig.CreateElement('system.webServer')
  $modulesNode = $appConfig.CreateElement('modules')
  $addNode = $appConfig.CreateElement('add')

  $attrib = $addNode.OwnerDocument.CreateAttribute('name')
  $addNode.Attributes.Append($attrib) | Out-File -Append log.txt
  $addNode.SetAttribute('name', 'ApplicationSecurityAuthenticationModule') | Out-File -Append log.txt

  $attrib = $addNode.OwnerDocument.CreateAttribute('type')
  $addNode.Attributes.Append($attrib) | Out-File -Append log.txt
  $addNode.SetAttribute('type', 'Advantys.Security.Http.AuthenticationModule') | Out-File -Append log.txt

  $modulesNode.AppendChild($addNode) | Out-File -Append log.txt
  $webServerNode.AppendChild($modulesNode) | Out-File -Append log.txt
  $appConfig.configuration.AppendChild($webServerNode) | Out-File -Append log.txt
}
else{
  if ($appConfig.configuration.SelectNodes('system.webServer').Count -ne 0){
    $appConfig.configuration.RemoveChild('system.webServer')
  }
}

$appConfig.Save($webConfig)
Write-Host "`r$message Done."

$message = "27. Modifying the services web.config..."
$message | Out-File -Append log.txt
Write-Host -NoNewLine $message
$esConfig = $settings.ServicesPath + "\bin\WfgWorkflowEngineService.exe.config"
$appConfig = [xml](cat $esConfig)
$foundSettings = $appConfig.configuration.appSettings.add | ? { $_.key -eq "WebConfigPath"}
if ($foundSettings.Count -ne 0) {
  $foundSettings | % { $_.value = $settings.WebAppPath + "\web.config" }
}
$appConfig.Save($esConfig)

$dsConfig = $settings.ServicesPath + "\bin\WfgDirectoriesSyncService.exe.config"
$appConfig = [xml](cat $dsConfig)
$foundSettings = $appConfig.configuration.appSettings.add | ? { $_.key -eq "WebConfigPath"}
if ($foundSettings.Count -ne 0) {
  $foundSettings | % { $_.value = $settings.WebAppPath + "\web.config" }
}
$appConfig.Save($dsConfig)

Write-Host "`r$message Done."

# Installing WFG's windows services (Engine Service and Directory Service)
if ((Get-WmiObject -Class Win32_Service -Filter "Name='WorkflowGenEngineService'") -eq $null){
  $message = "28. Installing the windows services..."
  $message | Out-File -Append log.txt
  Write-Host -NoNewLine $message
	$command = "'' | cmd.exe /C  '" + $settings.ServicesPath + "\bin\winsvc-install.cmd'"
	Invoke-Expression -Command $command | Out-File -Append log.txt
  Write-Host "`r$message Done."
}
else{
  $message = "28. Windows services already installed."
  $message | Out-File -Append log.txt
  Write-Host $message
	Start-Service "WorkflowGenEngineService"
  echo "    Starting WorkflowGen Engine Service."
	Start-Service "WorkflowGenDirSyncService"
  echo "    Starting WorkflowGen Directory Sync Service."
}

$iis = Get-Service -Name "w3svc"
if ($iis.status -ne "Running"){
	if ($Docker -eq $true){
		./ServiceMonitor.exe "w3svc"
	}
	else{
		Start-Service "w3svc"
	}
}

$WFGWebsite = (Get-WebSite $settings.IISSiteName)
if ($WFGWebsite -ne $null -And $WFGWebsite.State -eq "Stopped"){
  Write-Host -NoNewLine "29. Starting site:"  $settings.IISSiteName + "..."
  Start-WebSite $settings.IISSiteName
  Write-Host "`r$message Done."

}
if (Test-Path ".\tempconfig.json"){
  Remove-Item ".\tempconfig.json"
}
if (Test-Path ".\InstallUtil.InstallLog"){
  Remove-Item ".\InstallUtil.InstallLog"
}
Write-Host "`n"
Write-Host "Installation complete."
